local key = ModPath .. '	' .. RequiredScript
if _G[key] then return else _G[key] = true end

_G.Perfometer = {}
Perfometer.path = ModPath
Perfometer.data_path = SavePath .. 'perfometer.txt'
Perfometer.date = os.date('%Y-%m-%d_%H-%M')
Perfometer.colors = {
	body_collisions = '#bb9aed',
	corpses = '#49628a',
	delay = 'red',
	frames = '#0fb300',
	npcs = '#0000ff',
	tasks = '#ff730f',
}
Perfometer.style = {
	delay_weight = 0.1,
	font_size = '0.6em',
	grid_weight = 0.1,
	grid_step_x = 60, -- 1 per minute
	grid_step_y = 20,
	line_weight = 0.5,
	margin = 20,
	scale_ratio = 5,
}
Perfometer.settings = {
	full_team_balance = false,
	log_slow_locations = false,
	output_body_collisions = true,
	output_corpses = true,
	output_frames = true,
	output_npcs = true,
	output_task_delays = true,
	output_tasks = true,
}
Perfometer.difficulty_names = {
	easy = 'Easy',
	normal = 'Normal',
	hard = 'Hard',
	overkill = 'VeryHard',
	overkill_145 = 'Overkill',
	easy_wish = 'Mayhem',
	overkill_290 = 'DeathWish',
	sm_wish = 'DeathSentence',
}

function Perfometer:load()
	local fh = io.open(self.data_path, 'r')
	if fh then
		for k, v in pairs(json.decode(fh:read('*all'))) do
			self.settings[k] = v
		end
		fh:close()
	end
end

function Perfometer:save()
	local fh = io.open(self.data_path, 'w+')
	if fh then
		fh:write(json.encode(self.settings))
		fh:close()
	end
end

function Perfometer:init_data()
	self.body_collisions = {}
	self.corpses = {}
	self.frames = {}
	self.npcs = {}
	self.tasks = {}
	self.avg_delays = {}
	self.max_delays = {}
end

function Perfometer:get_filename()
	local filename = ('%sresults/perfs_payday2_%s_%s_%s_%ix%i_%iHz_%iocc.svg'):format(
		self.path,
		self.date,
		(Global.game_settings.level_id:gsub('_', '')),
		self.difficulty_names[Global.game_settings.difficulty] or Global.game_settings.difficulty,
		RenderSettings.resolution.x,
		RenderSettings.resolution.y,
		RenderSettings.resolution.z,
		World:occlusion_manager():max_occluder_tests()
	)
	return filename
end

function Perfometer.line(color, x1, x2, y1, y2, width, dash)
	return ('<line stroke="%s" stroke-width="%.1f" x1="%i" x2="%i" y1="%i" y2="%i"%s></line>\n'):format(
		color,
		width or 1,
		x1,
		x2,
		y1,
		y2,
		dash and (' stroke-dasharray="' .. dash .. '"') or ''
	)
end

function Perfometer.polyline(tbl, color, width)
	local v_max = tbl[1]
	local v0, v1, t0, t1
	local tmp = {}
	for t2, v2 in ipairs(tbl) do
		v_max = math.max(v2, v_max)
		if v0 and (v2 - v1) / (t2 - t1) == (v1 - v0) / (t1 - t0) then
			table.remove(tmp)
		end
		table.insert(tmp, t2 .. ',' .. (v2 ~= 0 and '-' or '') .. v2)
		v0 = v1
		v1 = v2
		t0 = t1
		t1 = t2
	end

	local result = '<polyline fill="none" stroke="' .. color .. '" stroke-width="' .. width .. '" points="' .. table.concat(tmp, '\n') .. '"/>\n'
	return result, v_max
end

function Perfometer.get_max_value(tbl)
	local result = 0
	for _, v in ipairs(tbl) do
		result = math.max(result, v)
	end
	return result
end

function Perfometer:export(white_background)
	if #self.frames < 10 then
		return false
	end

	local filename = self:get_filename()
	local fh = io.open(filename, 'w+')
	if not fh then
		return false
	end

	local result = pcall(function()
		self:_export(fh, white_background)
	end)

	fh:close()

	return result and filename
end

function Perfometer:_export(fh, white_background)
	local width = 1 + #self.frames
	local legend = { left = {}, right = {} }
	local data = { lines = {}, maxima = {} }

	local function stack_legend(side, caption, color)
		local on_left = side == 'left'
		local legend_side = on_left and legend.left or legend.right
		local text = ('<text fill="%s" font-size="%s" x="%i" y="%i" text-anchor="%s">%s</text>\n'):format(
			color,
			self.style.font_size,
			on_left and 0 or width,
			self.style.margin * (1 + #legend_side * 0.5),
			on_left and 'start' or 'end',
			caption
		)
		table.insert(legend_side, text)
	end

	local function store_polyline(values, color)
		local line, max = self.polyline(values, color, self.style.line_weight)
		table.insert(data.lines, line)
		table.insert(data.maxima, max)
	end

	if self.settings.output_npcs then
		store_polyline(self.npcs, self.colors.npcs)
		stack_legend('left', 'Amount of NPC (cops + civilians)', self.colors.npcs)
	end

	if self.settings.output_frames then
		store_polyline(self.frames, self.colors.frames)
		stack_legend('left', 'Frames per second', self.colors.frames)
	end

	if self.settings.output_corpses then
		local scaled_corpses, side
		if self.get_max_value(self.corpses) <= math.max(0, unpack(data.maxima)) then
			side = 'left'
			scaled_corpses = self.corpses
		else
			side = 'right'
			scaled_corpses = {}
			for t, v in ipairs(self.corpses) do
				scaled_corpses[t] = v / self.style.scale_ratio
			end
		end
		store_polyline(scaled_corpses, self.colors.corpses)
		stack_legend(side, 'Amount of corpses', self.colors.corpses)
	end

	if self.settings.output_body_collisions then
		local scaled_body_collisions = {}
		for t, v in ipairs(self.body_collisions) do
			scaled_body_collisions[t] = v / self.style.scale_ratio
		end
		store_polyline(scaled_body_collisions, self.colors.body_collisions)
		stack_legend('right', 'Amount of body collisions', self.colors.body_collisions)
	end

	if self.settings.output_tasks then
		local scaled_tasks = {}
		for t, v in ipairs(self.tasks) do
			scaled_tasks[t] = v / self.style.scale_ratio
		end
		store_polyline(scaled_tasks, self.colors.tasks)
		stack_legend('right', 'Tasks executed per second', self.colors.tasks)
	end

	if self.settings.output_task_delays then
		stack_legend('right', 'Delay (avg to max in sec/100)', self.colors.delay)
	end

	local height = 1 + math.max(unpack(data.maxima))

	local document_x = -self.style.margin
	local document_y = -height - 7
	fh:write(('<svg viewBox="%i %i %i %i" version="1.1" xmlns="http://www.w3.org/2000/svg">\n'):format(
		document_x,
		document_y,
		width + 2 * self.style.margin,
		height + (1 + math.max(#legend.left, #legend.right) * 0.5) * self.style.margin
	))

	if white_background then
		fh:write(('<rect x="%i" y="%i" width="100%%" height="100%%" fill="white"/>\n'):format(
			document_x,
			document_y
		))
	end

	if self.settings.output_task_delays then
		local delay_scale = self.style.scale_ratio * -4 -- to get 100th of second and use right y-axis
		for t, max in ipairs(self.max_delays) do
			if max * delay_scale <= -1 then
				local min = self.avg_delays[t]
				fh:write(self.line(self.colors.delay, t, t, min * delay_scale, max * delay_scale, self.style.delay_weight))
			end
		end
	end

	for _, text in ipairs(data.lines) do
		fh:write(text)
	end
	for _, text in ipairs(legend.left) do
		fh:write(text)
	end
	for _, text in ipairs(legend.right) do
		fh:write(text)
	end

	for i = 0, width, self.style.grid_step_x do
		fh:write(('<text font-size="%s" x="%i" y="10">%i</text>\n'):format(
			self.style.font_size,
			i,
			i
		))
		fh:write(self.line('grey', i, i, -height, 0, self.style.grid_weight, '1 2'))
	end

	for i = 0, height, self.style.grid_step_y do
		fh:write(('<text font-size="%s" x="-%i" y="-%i" text-anchor="end">%i</text>\n'):format(
			self.style.font_size,
			self.style.margin / 5,
			i,
			i
		))
		fh:write(('<text font-size="%s" fill="%s" x="%i" y="-%i">%i</text>\n'):format(
			self.style.font_size,
			self.colors.tasks,
			width,
			i,
			i * self.style.scale_ratio
		))
		fh:write(self.line('grey', 0, width, -i, -i, self.style.grid_weight, '1 2'))
	end

	fh:write(self.line('black', 0, 0, -height, 2))
	fh:write(self.line('black', -2, width, 0, 0))

	fh:write('</svg>')
end

function Perfometer:view_in_overlay()
	local filename = self:export(true)
	if filename then
		local url_svg = 'file:///' .. Application:base_path():gsub('\\', '/'):gsub(' ', '%%20') .. filename
		Steam:overlay_activate('url', url_svg)
		-- Steam:overlay_activate('url', 'data:text/html;base64,' .. require('lib/utils/base64').encode('<html><body style="background-color: white"><img src="' .. url_svg .. '"></body></html>'))
	end
end

Hooks:Add('MenuManagerInitialize', 'MenuManagerInitialize_Perfometer', function(menu_manager)
	MenuCallbackHandler.PerfometerSave = function()
		Perfometer:save()
	end

	MenuCallbackHandler.PerfometerMenuCheckboxClbk = function(this, item)
		Perfometer.settings[item:name()] = item:value() == 'on'
	end

	MenuCallbackHandler.PerfometerMenuViewInOverlayClbk = function(this, item)
		Perfometer:view_in_overlay()
	end

	Perfometer:load()
	MenuHelper:LoadFromJsonFile(Perfometer.path .. 'menu/options.txt', Perfometer, Perfometer.settings)
end)

Hooks:Add('LocalizationManagerPostInit', 'LocalizationManagerPostInit_Perfometer', function(loc)
	local language_filename

	for _, filename in pairs(file.GetFiles(Perfometer.path .. 'loc/')) do
		local str = filename:match('^(.*).txt$')
		if str and Idstring(str) and Idstring(str):key() == SystemInfo:language():key() then
			language_filename = filename
			break
		end
	end

	if language_filename then
		loc:load_localization_file(Perfometer.path .. 'loc/' .. language_filename)
	end
	loc:load_localization_file(Perfometer.path .. 'loc/english.txt', false)
end)
